import 'dart:async';
import 'package:flutter/material.dart';
import 'package:flutter_easyrefresh/easy_refresh.dart';
import 'package:get/get.dart';

import 'loader//pq_pre_loader.dart';
import 'status_widgets/pq_no_more_data_view.dart';
import 'pq_list_controller.dart';

typedef ListItemBuilder = Widget Function(BuildContext context, int index);

class PqListView<T extends DisposableInterface> extends StatefulWidget {
  PqListView({
    Key? key,
    required this.controller,
    required this.itemBuilder,
    required this.itemCount,
    this.separatorBuilder,
    EasyRefreshController? refreshController,
    this.onRefresh,
    this.onLoadMore,
    this.scrollController,
    this.header,
    this.footer,
    this.firstRefresh = false,
    this.padding,
    this.topBouncing = true,
    this.bottomBouncing = true,
    this.preloading = false,
    this.preloadingDistanceLimit = PqPreLoader.DEFAULT_PRELOAD_DISTANCE_LIMIT,
    this.autoAddNoMoreView = true,
    this.noMoreDataView,
  })  : this.refreshController = refreshController ?? EasyRefreshController(),
        super(key: key);

  /// GetController，这里建议传入[PqListController]，当controller是YhListController
  /// 会根据nextId自行处理是否显示上拉加载更多功能
  final T controller;

  /// 列表Cell的构建
  final ListItemBuilder itemBuilder;

  /// 列表分割线的构建
  final ListItemBuilder? separatorBuilder;

  /// 列表的长度
  final int itemCount;

  /// 刷新controller
  final EasyRefreshController refreshController;

  /// 刷新
  final Future<void> Function()? onRefresh;

  /// 加载更多
  final Future<void> Function()? onLoadMore;

  /// 滚动controller
  final ScrollController? scrollController;

  /// 刷新头
  final Header? header;

  /// 刷新脚
  final Footer? footer;

  /// 是否开启首次刷新
  final bool firstRefresh;

  /// List view padding
  final EdgeInsetsGeometry? padding;

  /// 顶部回弹(Header的overScroll属性优先，且onRefresh和header都为null时生效)
  final bool topBouncing;

  /// 底部回弹(Footer的overScroll属性优先，且onLoad和footer都为null时生效)
  final bool bottomBouncing;

  /// 是否开启预加载 默认是关闭状态
  final bool preloading;

  /// 预加载高度 默认是[PqPreLoader.DEFAULT_PRELOAD_DISTANCE_LIMIT]
  final double preloadingDistanceLimit;

  /// 自动添加没有更多视图 [PqNoMoreDataView]
  final bool autoAddNoMoreView;

  /// 没有更多视图的View
  final Widget? noMoreDataView;

  @override
  _PqListViewState createState() => _PqListViewState();
}

class _PqListViewState extends State<PqListView> {
  ScrollController? scrollController;
  PqPreLoader? _preLoader;

  @override
  initState() {
    super.initState();
    scrollController = widget.scrollController;
    if (widget.controller is PqListController) {
      (widget.controller as PqListController).refreshController = widget.refreshController;
    }
  }

  @override
  void dispose() {
    _subscription?.cancel();
    _loadingController?.close();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    if (widget.preloading && widget.scrollController == null) {
      scrollController = PrimaryScrollController.of(context);
      _preLoader ??=
          PqPreLoader(scrollController!, loadMoreListener: _sendLoading);
      _singleLoadMore();
    }
    return Obx(() {
      bool noMore = false;
      bool isListController = widget.controller is PqListController;
      if (isListController) {
        noMore = (widget.controller as PqListController).isLastPage.value;
      }

      bool showNoMoreDataView = false;
      if (widget.autoAddNoMoreView &&
          (isListController) &&
          (widget.controller as PqListController).isLastPage.value) {
        showNoMoreDataView = true;
      }

      Widget Function(BuildContext context, int index) _itemBuilder =
          (BuildContext context, int index) {
        if (index == widget.itemCount && showNoMoreDataView) {
          return widget.noMoreDataView ?? PqNoMoreDataView();
        }

        return widget.itemBuilder(context, index);
      };

      return EasyRefresh(
        controller: widget.refreshController,
        onRefresh: isListController
            ? (widget.controller as PqListController).onRefresh
            : widget.onRefresh,
        onLoad: noMore
            ? null
            : () async {
                if (widget.preloading) {
                  _sendLoading();
                } else {
                  await _onLoadMore();
                }
              },
        scrollController: scrollController,
        header: widget.header,
        footer: widget.footer,
        firstRefresh: widget.firstRefresh,
        topBouncing: widget.topBouncing,
        bottomBouncing: widget.bottomBouncing,
        child: ListView.separated(
          padding: widget.padding,
          itemBuilder: _itemBuilder,
          separatorBuilder:
              widget.separatorBuilder ?? (_, __) => SizedBox.shrink(),
          itemCount: (showNoMoreDataView && widget.itemCount > 0)
              ? widget.itemCount + 1
              : widget.itemCount,
        ),
      );
    });
  }

  _sendLoading() {
    if (_loading) return;
    _loadingController?.add(_loading);
  }

  var _loading = false;
  GetStream? _loadingController;
  LightSubscription<dynamic>? _subscription;

  _singleLoadMore() {
    _loadingController ??= GetStream();
    _subscription ??= _loadingController!.listen((event) async {
      if (_loading) return;
      _loading = true;
      await _onLoadMore();
      _loading = false;
    });
  }

  Future<void> _onLoadMore() async {
    return await (widget.controller is PqListController
        ? (widget.controller as PqListController).onLoadMore()
        : widget.onLoadMore?.call());
  }
}
